'****************************************************************
'*  Name    : RTC_P+.Inc                                        *
'*  Author  : John Barrat                                       *
'*  Notice  : Copyright (c) 2005 John Barrat                    *
'*          : All Rights Reserved                               *
'*  Date    : 18/4/2005                                        *
'*  Version : 1.0                                               *
'*  Notes   : Generalised Subroutines for use with DS1307 clock *
'*          : Requires the following Includes:                  *
'*          :   LCD_INC.bas                                     *
'*          :   CON_INC.bas                                     *
'****************************************************************
' The following subroutines provide access to RTC functions using the DS1307 clock chip

' Input/Output Var types
'   RTC_Cont    Boolean
'   RTC_Data    Byte
'   RTCArry     Byte Array  (Alias RTCSec, RTCMin, RTCHour, RTCWDay, RTCDay, RTCMonth, RTCYear)
'   SetArry     Byte Array  (Alias SetSec, SetMin, SetHour, SetWDay, SetDay, SetMonth, SetYear)

' Routine       Input Var   Output Var  Function
'----------------------------------------------------------------------------------------------- 
' CheckRTC      None        RTC_Cont    Returns False if RTC clock not set up
'                                       Presets starting value in the RTC Array if not set up

' RTC_Init      None        RTC_Cont    Retrieves the DS Flag and set the DS indicatore
'                                       Adds the RTC tasks to task list and Returns True if OK

' TSetUp        None        None        Converts BCD values in RTCArry to BIN in SetRTCArry
'                                       Sets the DS indicators in SetArray

' ConvBCD       Passed     Returned     Converts Passed BIN Value to BCD

' ConvBIN		Passed	   Returned     Converts Passed BCD Value to Binary

' FindDays      SetMonth    RTC_Data    Returns the number of days in the month taking into 
'               SetYear                 account leap years

' DispTick		tkRow                   Displays at position tkRow/tkCol the current state of
'  				tkCol                   the tick indicator
'				TickInd  				

' SetDS_Ind     SetDSTime   SetDSInd    Sets SetDS Indicator to the state defined by SetDSTime

' DispTimeDate  RTCArry     None        Displays Time/Date in RTCArry formatted HH:MM* DD-MMM-YY

' DispTimeDate2 SetRTCArry  None        As DispTimeDate but using data held in SetRTCArry

' DispDate2     SetRTCArry  None        Displays Date in SetRTCArry formatted DD-MM-YY
' DispTime2     SetRTCArry  None        Displays Time in SetRTCArry formatted HH:MM

' ToggleDS      DSTime      DSTime      Toggles Daylight Saving states on/off, updates SetDSInd, 
'                           DSInd       adjusts the current time +/- 1 hour
'                           RTCHour     

' Next 5 subroutines inc/dec timedate - RTC_Cont controls direction True = Inc, False = Dec
' StepHour      RTC_Cont                Inc/Dec Hours rolling over at 23 Hours
'               SetHour     SetHour
' StepMin       RTC_Cont                Inc/Dec Mins rolling over at 59 Mins
'               SetMin      SetMin      
' StepDay       RTC_Cont                Inc/Dec Days rolling over at days in current month    
'               SetDay      SetDay
'               SetMonth
' StepMonth     RTC_Cont                Inc/Dec Month rolling over at 12 
'               SetMonth    SetMonth
'               SetDay      SetDay      SetDays adjusted where necessary
' StepYear      RTC_Cont    
'               SetYear     SetYear     Inc/Dec Year rolling over at 99
'               SetMonth
'               SetDay      SetDay      SetDays adjusted where necessary

' RTCUpDate     SetRTCArry  RTCArry     Calculates workday number, updates the RTCArry and RTC

' SaveRTC       RTCArry     None        Updates the RTC chip from the RTCArry

' The DS1307 also contains 56 bytes of battery backed RAM.  The first location (Loc 0) is 
' reserved as a status byte and bit 0 is reserved for the Daylight Saving status.
' The RAM can be accessed through following:
' SaveNVRam     RTC_Data    None        RTC_Data holds the value to be written and NV_Addr holds
'               NV_Addr     None        the location to save it to. RTC_Cont returns false if
'                           RTC_Cont    the Save Fails
' ReadNVRam     NV_Addr     RTC_Data    The data held in location NVAddr is returned in RTC_Data 
'                           RTC_Cont    RTC returns false if read fails
'-----------------------------------------------------------------------------------------------
'Device = 18F252
' RTC Constants
Symbol RTCTickOn   = ":"      ' Tick indicator character
Symbol RTCSumOn    = "*"      ' Daylight Saving indicator character
Symbol RTCIndOff   = " "      ' Indicator off character (Normally a blank)
Symbol RTCDevOut   = $D0      ' Dev Address Write of DS1307 RTC, RTC 00H to 07H, Memory are 08H to 3FH
Symbol RTCDevIn    = $D1      ' Dev Address Read of DS1307 RTC, RTC 00H to 07H, Memory are 08H to 3FH
Symbol RTCSQWaveOn = $10      ' Control Byte to set 50Hz Square wave on RTC
Symbol RTC_Addr    = $00
Symbol NV_OffSet   = $08

' Public Variables
Dim RTCArry[7] As Byte         ' RTC Array
Symbol RTCSec   RTCArry#0      ' Seconds
Symbol RTCMin   RTCArry#1      ' Minutes
Symbol RTCHour  RTCArry#2      ' Hours
Symbol RTCWDay  RTCArry#3      ' Weekday
Symbol RTCDay   RTCArry#4      ' Day
Symbol RTCMonth RTCArry#5      ' Months
Symbol RTCYear  RTCArry#6      ' Year
Dim TickInd     As Byte    	   ' This Var holds the flashing ":" on the Time display
Dim DSInd       As Byte   	   ' Holds the char to be displayed "Daylight Saving"/"Normal"

Dim DSTime      As Bit         ' True(1) indicates Daylight Saving Time
Dim TkRow 		As Byte        ' Row on which Tick Indicator is positioned
Dim TkCol 	    As Byte		   ' Character position of Tick Indicator
Dim SecCtr 		As Byte
Dim SetRTCArry[7]  As Byte     ' Editable RTC Array
Symbol SetSec   SetRTCArry#0   ' Seconds
Symbol SetMin   SetRTCArry#1   ' Minutes
Symbol SetHour  SetRTCArry#2   ' Hours
Symbol SetWDay  SetRTCArry#3   ' Weekday
Symbol SetDay   SetRTCArry#4   ' Day
Symbol SetMonth SetRTCArry#5   ' Months
Symbol SetYear  SetRTCArry#6   ' Year
Dim SetDSInd    As Byte        ' Holds the char to indicate Daylight Saving 
Dim SetDSTime   As Bit         ' True(1) indicates Daylight Saving Time

Dim RTC_Cont    As Bit         ' Used to pass a boolean to a function and return a boolean result
Dim RTC_Data    As Byte        ' Used to receive data into a function and return result
Dim NV_Addr     As Byte        ' Address into the 56 byte NV Ram on the DS1307

' Internal Variables
Dim RTC_IntA    As Byte         ' Only used internally
Dim RTC_IntB    As Byte         ' Only used internally   
Dim RTC_IntC  	As Byte         ' Only used internally
Dim RTC_IntW1 	As Word         ' Only used internally
Dim RTC_IntW2 	As Word         ' Only used internally
Dim DateChanged As Bit          ' True when date has changed as a result of reading the RTC 
Dim PP4 As Byte SYSTEM          ' Used to hold the state of the I2C Ack return
Symbol I2CErr   PP4.0           ' I2C Fail Flag

GoTo Skip_RTC_Inc

' Converts the current RTC BCD Values back into binary values 
' and leaves them in SetHour, SetMin, SetYear, SetMonth, SetDay ready for editing
' and the Daylight saving indicator set in SetDSTime
SMALL_MICRO_MODEL = On
TSetup:
    For RTC_IntA = 0 To 6
        GoSub ConvBIN [RTCArry[RTC_IntA]], SetRTCArry[RTC_IntA]
    Next
    SetDSTime = DSTime
    GoSub Set_DSInd
Return

ConvBIN:
' Function to convert from BCD to Bin - Syntax GoSub ConvBIN [BCD as Byte], Bin as Byte
' Call with Value to be converted and returns with converted value
    Pop RTC_IntB
	RTC_Data = RTC_IntB >> 4
	RTC_Data = RTC_Data * 10
	RTC_IntB = RTC_IntB & $0F
	RTC_Data = RTC_Data + RTC_IntB
Return RTC_Data

ConvBCD:
' Function to convert from Bin to BCD - Syntax GoSub ConvBCD [Bin as Byte], BCD as Byte
' Called with Value to convert and returns with converted value
    Pop RTC_Data
	RTC_IntB = Dig RTC_Data, 1
	RTC_IntB = RTC_IntB << 4
	RTC_IntB = RTC_IntB + Dig RTC_Data, 0
Return RTC_IntB

' Subroutine works out Number of Days in the Month taking the data from the RTCSetArry
' exits with number of days in RTC_Data 
FindDays:
    ' Get the 'usual' days of the Month
    RTC_Data = LookUp SetMonth,[0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    ' adjust for Leap-Year
    If SetMonth = 2 Then      'SetMonth = February then   (SetRTCArry[5] )
       If SetYear & $03 = 0 Then RTC_Data = 29
    EndIf
Return

DispTick:
' Routine displays the current state of the tick indicator at position tkRow, TkCol
    If SQWPin = 0 Then TickInd = RTCTickOn: Else: TickInd = RTCIndOff
	Print At TkRow, TkCol, TickInd
Return
	 
'   Subroutines to display or return time & date in the form:  
'   HH:MM* DD-MMM-YY  16 chars on a single line  
'   Using the RTC Values
DispTimeDate:			' Displays Time Date start of row defined by TkRow, Sets TkCol
    Cursor TkRow, 1	    ' Set cursor to start of row on which tick indicator will be displayed
    TkCol = 3
    ' Displays Time/Date from current cursor position - must be start of 16 char line
    Clear MsgArry
    GoSub StrTimeDate
    GoTo PrintStr

' Returns a time & date string in MsgArry in the form HH:MM* DD-MMM-YY using RTC Values
StrTimeDate:
    GoSub StrTime
    MsgArry = MsgArry + DSInd + " "
StrDate:
    MsgArry = MsgArry + Str$(HEX2 RTCDay) + "-"              ' Days - separated by Space
    GoSub ConvBIN [RTCMonth], RTC_Data                       ' Get Month in Binary
    GoSub StrMonth                                           ' Get it in "mmm" format
    MsgArry = MsgArry + "-" + Str$(HEX2 RTCYear)             ' Just show last 2 digits of year
Return

' As DispTimeDate but using the Set Values as source
DispTimeDate2: 
    GoSub StrTimeDate2
    GoTo PrintStr
' Returns a time & date string in MsgArry in the form HH:MM* DD-MMM-YY using SetArry values 
StrTimeDate2:
    GoSub StrTime2
    MsgArry = MsgArry + SetDSInd + " "
StrDate2:
    MsgArry = MsgArry + Str$(DEC2 SetDay) + "-"              ' Days - separated by "-"
    RTC_Data = SetMonth: GoSub StrMonth                      ' Display month in "mmm" format
    MsgArry = MsgArry + "-" + Str$(DEC2 SetYear)             ' Just show last 2 digits of year
Return

' Displays Time in the form HH:MM* using RTC values, 
' Positioned around location of Tick Indicator (TkRow, TkCol)
DispTime:
    GoSub StrTime 	 	 	  		 	   	   		 		 ' Build the Time Message string
    Cursor TkRow, TkCol - 2                                  ' Set Cursor to start of Time 
    GoTo PrintStr
StrTime:
' Formats a string in the form HH:MM from the BCD coded RTC values and returns it in MsgArry
    MsgArry = Str$(HEX2 RTCHour) + ":" + Str$(HEX2 RTCMin)
Return

DispTime2:
    Clear MsgArry
    GoSub StrTime2
    GoTo PrintStr
StrTime2:
    ' Formats a string in the form "HH:MM" using the SetTime values and adds it into the MsgArry
    MsgArry = Str$ (DEC2 SetHour)  + ":" + Str$ (DEC2 SetMin)
Return

DispDayName:   '   Syntax GoSub DisDayName [DayNum] where Monday = 1, Sunday = 7
'   Displays the day name in the form DDD using value passed 
    Clear MsgArry
    GoSub StrDayName
    GoTo PrintStr

StrDayName:
'   Returns the Day of the Week in "DDD" format 
    Pop RTC_Data
    RTC_Data = (RTC_Data * 3)
	RTC_Data = RTC_Data - 2
    MsgArry = MsgArry + Mid$(DayNames, RTC_Data, 3)	
Return

' Subroutine to Display Month held as MM in RTC_Data as MMM at Current LCD Cursor Position
StrMonth:
    RTC_Data = (RTC_Data * 3)
	RTC_Data = RTC_Data - 2               
    MsgArry = MsgArry + Mid$(MthNames, RTC_Data, 3)	
Return

StrRTCUnit: ' Syntax GoSub StrRTCUnit [Index]
' Returns a fixed length string of 5 chars referenced by the index Syntax 
' for the following Index values:
' 1 returns "Secs ", 2 returns "Hours", 3 returns "Mins "
' 4 Returns "Day  ", 5 Returns "Month", 6 Returns "Year "
    Pop RTC_Data
    Inc RTC_Data  ' Skip secs 
    RTC_Data = (RTC_Data * 5)
	RTC_Data = RTC_Data - 4
    MsgArry = MsgArry + Mid$(RTCUnits, RTC_Data, 5)
Return

PrintStr:
    Print MsgArry
    Clear MsgArry
Return

' Subroutine sets value of the Set Summer time indicator to reflect the state of SetDSTime
Set_DSInd:
    If SetDSTime = True Then                ' Set the SetDSIndicator
        SetDSInd = RTCSumOn                 ' Set Summer Time Indicator
    Else
        SetDSInd = RTCIndOff                ' Set Winter Time Indicator
    EndIf
Return

' Subroutine to switch between Winter & Summer time adjusts clock forward and back 1 hour
' Amends Hour value held in SetHour - Operates on the SetArray - Call TSetup to prime
ToggleDS:
    SetDSTime = ~SetDSTime                  ' Toggle Set SummerTime flag
    GoSub Set_DSInd                         ' Set the Indicator
    RTC_Cont = SetDSTime                    ' Set Direction and Adjust hour +/- 1 
    
' Steps SetHour forward or back resetting at 23 to 0 - RTC_Cont = True is Plus 1 hour
StepHour:                                   
    If RTC_Cont = True Then                 ' RTC_Cont define direction 1 = Forward, 0 = Back
		Inc SetHour                         ' Increment the Hours
		If SetHour > 23 Then SetHour = 0    ' Reset to 0 if overflowed
    Else
		If SetHour = 0 Then SetHour = 24    ' Reset to 24 if at 0
        Dec SetHour                        	' Decrement 1 hour
    EndIf
Return
    
' Steps SetMin forward or back resetting at 59 to 0 - RTC_Cont = True is Plus 1 minute
StepMin:                                    
    If RTC_Cont = True Then                 ' RTC_Cont define direction 1 = Forward, 0 = Back
        Inc SetMin                          ' Increment the Minutes and check
        If SetMin > 59 Then SetMin = 0      ' if they have reached 60 if so reset to 0
    Else
	    If SetMin = 0 Then SetMin = 60      ' Reset to 60 if at 0
	    Dec SetMin                          ' Decrement 1 minute
    EndIf
Return

' Steps SetDay forward or back resetting at Number of days in current month
' RTC_Cont = True is Plus 1 day
StepDay:
    GoSub FindDays                          ' Check number of days in the current month
    If RTC_Cont = True Then                 ' RTC_Cont define direction 1 = Next, 0 = Previous
        Inc SetDay                          ' Increment the current day and check if it exceeds
        If SetDay > RTC_Data Then SetDay = 1' number of days in month if so reset to 1
    Else
        If SetDay = 0 Then SetDay = RTC_Data + 1 ' Reset to Days in month + 1
        Dec SetDay                          ' and decrement day
    EndIf
Return

' Steps SetMonth forward or back resetting at 12 back to 1 - adjust the SetDay if not valid for
' current month - RTC_Cont = True is Plus 1 Month
StepMonth:
    If RTC_Cont = True Then                 ' RTC_Cont defines direction 1 = Next, 0 = Previous
        Inc SetMonth                        ' Increment the current day and check
        If SetMonth > 12 Then SetMonth = 1  ' for over-run, reset to 1 if necessary
    Else
    	If SetMonth = 1 Then SetMonth = 13  ' If month at 1 Reset to 13
    	Dec SetMonth                        ' and decrement month                            
    EndIf
    GoTo ChkValidDate

' Steps SetYear forward or back resetting at 99 back to 00 - adjust the SetDay in not valid for
' current month - RTC_Cont = True is Plus 1 Year
StepYear:
    If RTC_Cont = True Then                 ' RTC_Cont defines direction 1 = Next, 0 = Previous
        Inc SetYear                         ' Inc year 
        If SetYear > 99 Then SetYear = 0    ' if > 99 reset to 0
    Else
        If SetYear = 0 Then SetYear = 100   ' If 0 reset to 100
        Dec SetYear                         ' and decrement year
    EndIf

ChkTimeAndDate:                        		' Check Time is a valid time 
	RTC_Cont = False		                ' Set the return value to false (OK)
    If SetMin > 59 Then RTC_Cont = True     ' Mins > 59 - Error set return to True
    If SetHour > 23 Then RTC_Cont = True    ' Hours > 23 - Error set return to True
    If RTC_Cont = True Then Return
    If SetMonth > 12 Then RTC_Cont = True   ' Check month within boundary
    If SetMonth = 0 Then RTC_Cont = True
ChkValidDate:
    GoSub FindDays                          ' Check that Day still relevant to new month
    If SetDay > RTC_Data Then               ' If Day is outside current month
        SetDay = RTC_Data                   ' force day to last valid day of month
        RTC_Cont = True                     ' RTC_Cont True Indicates Day was changed
    Else
        RTC_Cont = False                    
    EndIf
Return

RTCUpDate:
'This routine calculates the Working day number from the Date set in the SetArray
'and save the new date/time settings in the RTC.

'   Melanie's fudge for calculating Days of Week using integer math...
    RTC_IntA = SetYear + 4: RTC_IntB = SetMonth
    If SetMonth < 3 Then
	    RTC_IntA = RTC_IntA - 1
	    RTC_IntB = RTC_IntB + 12
	EndIf
'    SetWDay = (SetDay+(153*RTC_IntB-457)/5+365*RTC_IntA+RTC_IntA/4-RTC_IntA/100+RTC_IntA/400+2)//7	
	RTC_IntW1 = 153 * RTC_IntB
	RTC_IntW1 = RTC_IntW1 - 457
	RTC_IntW1 = RTC_IntW1 / 5
	RTC_IntW2 = 365 * RTC_IntA
	RTC_IntW1 = RTC_IntW1 + RTC_IntW2
	
	RTC_IntB = RTC_IntA >> 2
	RTC_IntC = RTC_IntA / 100
	RTC_IntC = RTC_IntC + RTC_IntA / 400
	RTC_IntB = RTC_IntB - RTC_IntC
	
	RTC_IntW1 = SetDay + RTC_IntW1 + RTC_IntB + 2
	SetWDay = (RTC_IntW1)//7		
	SetWDay = SetWDay + 2
    'Save new time/date into RTC converting to BCD on the way
    RTC_IntA = $00
    GoSub I2CSetUp                        ' Prime RTC to point at RTC parameters
	Repeat
        GoSub ConvBCD [SetRTCArry[RTC_IntA]], RTC_IntB
        HBusOut RTC_IntB                  ' Send RTC Data to RTC and
        RTCArry[RTC_IntA] = RTC_IntB      ' put new value back into RTCarry
        If I2CErr = True Then GoTo NVErr  ' Check for NACK or missing Ack
		Inc RTC_IntA                      ' Increment pointer
	Until RTC_IntA = 7                    ' Repeat until we have reached end of array
    HBusOut RTCSQWaveOn
    HBStop
    ' Set the DSInd character and DSTime Flag to reflect the new values
    DSTime = SetDSTime
    DSInd  = SetDSInd
    NV_Addr = $00: GoSub ReadNVRam        ' Read NVRam (LSB of first byte contains DSTime state)
    If RTC_Cont = False Then Return       ' Errored reading NVRam so return
    RTC_Data.0 = DSTime                   ' Set to new value of DSTime and save the DSTime state   
SaveNVRam:
' This routine Saves a single byte into the 56 byte NVRAM - NV_Addr defines the locations $00-$37
' Data is returned in RTC_Data
    RTC_IntA = NV_Addr + NV_OffSet
    GoSub I2CSetUp
    HBusOut RTC_Data
    If I2CErr = True Then GoTo NVErr   
    HBStop
Return

ReadNVRam:
' This routine Reads a single byte from the 56 byte NVRAM - NVAddr defines the locations $00-$37
' Data is passed in RTC_Data - RTC_Cont Returns True OK, False Error
    RTC_IntA = NV_Addr + NV_OffSet
    GoSub I2CSetUp
    HBREStart
    HBusOut RTCDevIn
    HBusIn RTC_Data
    If I2CErr = True Then NVErr 
    HBStop    
Return
' Error so shut down Bus and return with the RTC_Cont bit set false
NVErr:
    RTC_Cont = False
    HBStop
Return

' Wakes up the RTC I2C device and sets its register pointer to the register defined in RTC_IntA
I2CSetUp:
    RTC_Cont = True
    HBStart
    HBusOut RTCDevOut
    HBusOut RTC_IntA
Return

' Should be called once per second
UpDateSeconds:
' Increments the seconds count resetting to 0 on the minute.
    Inc RTCSec
    If RTCSec & $0F > $9 Then 
	   RTCSec = RTCSec & $70
       RTCSec = RTCSec + $10
    EndIf  	   
    If RTCSec = $60 Then RTCSec = 00
Return

GetTimeDate:
' Reads RTC and returns the values into the RTC array and sets the DateChanged Flag if new date.
    RTC_IntA = RTCDay                   	  	 	 	' Save current day
	HBusIn RTCDevIn, $00, [Str RTCArry]
	If RTCDay <> RTC_IntA Then DateChanged = True    	' Set the new day flag if date changed
Return

CheckRTC:
' This routine checks the RTC to see if it has been previusly set up
'   If not set it will prime RTC Array with starting values and return False in RTC_Cont
'   If set up it will return RTC_Cont with True
	GoSub GetTimeDate
'   Decide if RTC Setup Required
	If RTCSec.7 = 1 Then        ' When DS1307 powers-up from cold Seconds bit 7 always high
        ' Default setting preset to the moment this program first born
        RTCSec   = $00: RTCMin = $00                    ' Preset Minutes & Seconds preset to 00
        RTCHour  = $12		                            ' Hours preset to 12'00, 24HR Mode
        RTCWDay  = $01		                            ' Weekday preset to 01
        RTCDay   = $01: RTCMonth = $06: RTCYear = $05   ' Preset Date to 01/06/2005
        DSTime   = True                                 ' Set to SummerTime
        DSInd    = RTCSumOn                             ' Set SummerTime Indicator char
        TickInd  = RTCTickOn                            ' Set Tick Indicator Char
        RTC_Cont = False                                ' Return with RTC_Cont not set
    Else                                                ' Previously set up so
        RTC_Cont = True                                 ' return with RTC_Cont Set
    EndIf
Return

' RTC_Init should be called after setting the time/date for first time to initialise the  
' Real Time Clock tasks into the task handler.
RTC_Init:
    RTC_Data = 0: RTC_Cont = 0            ' Reset External Variables  
    RTC_IntA = 0: RTC_IntB = 0            ' Reset Internal Variables
' Prime the DS Indicator
    NV_Addr = $00: GoSub ReadNVRam        ' Read NVRam (LSB of first byte contains DSTime state)
    If RTC_Cont = False Then Return       ' Errored reading NVRam so return
    DSTime = RTC_Data.0                   ' Get saved state of DSTime
    If DSTime = True Then 
        DSInd = RTCSumOn                  ' Set Summer time display character
    Else
        DSInd = RTCIndOff                 ' Clear Summer time Display charater
    EndIf
    TickInd = RTCTickOn                   ' Set Tick Indicator Char
	ADDTask UpDateSeconds, Rate2          ' Add the UpdateSecondes task to the 1/sec task list.
    RTC_Cont = True                       ' Return with INIT OK
Return

MthNames: CData "JanFebMarAprMayJunJulAugSepOctNovDec"
DayNames: CData "MonTueWedThuFriSatSun "
RTCUnits: CData "Secs HoursMins Day  MonthYear "

SMALL_MICRO_MODEL = Off

Skip_RTC_Inc:
